@@ -57,6 +57,14 @@
  * @CHAFA_COLOR_SPACE_MAX: Last supported color space plus one.
  **/
 
+/**
+ * ChafaDitherMode:
+ * @CHAFA_DITHER_MODE_NONE: No dithering.
+ * @CHAFA_DITHER_MODE_ORDERED: Ordered dithering (Bayer or similar).
+ * @CHAFA_DITHER_MODE_DIFFUSION: Error diffusion dithering (Floyd-Steinberg or similar).
+ * @CHAFA_DITHER_MODE_MAX: Last supported dither mode plus one.
+ **/
+
 /* Private */
 
 void
@@ -71,6 +79,10 @@ chafa_canvas_config_init (ChafaCanvasConfig *canvas_config)
     canvas_config->color_space = CHAFA_COLOR_SPACE_RGB;
     canvas_config->width = 80;
     canvas_config->height = 24;
+    canvas_config->dither_mode = CHAFA_DITHER_MODE_NONE;
+    canvas_config->dither_grain_width = 4;
+    canvas_config->dither_grain_height = 4;
+    canvas_config->dither_intensity = 1.0;
     canvas_config->fg_color_packed_rgb = 0xffffff;
     canvas_config->bg_color_packed_rgb = 0x000000;
     canvas_config->alpha_threshold = 127;
@@ -548,3 +560,130 @@ chafa_canvas_config_set_preprocessing_enabled (ChafaCanvasConfig *config, gboole
 
     config->preprocessing_enabled = preprocessing_enabled;
 }
+
+/**
+ * chafa_canvas_config_get_dither_mode:
+ * @config: A #ChafaCanvasConfig
+ *
+ * Returns @config's #ChafaDitherMode.
+ *
+ * Returns: The #ChafaDitherMode.
+ *
+ * Since: 1.2
+ **/
+ChafaDitherMode
+chafa_canvas_config_get_dither_mode (const ChafaCanvasConfig *config)
+{
+    g_return_val_if_fail (config != NULL, CHAFA_DITHER_MODE_NONE);
+    g_return_val_if_fail (config->refs > 0, CHAFA_DITHER_MODE_NONE);
+
+    return config->dither_mode;
+}
+
+/**
+ * chafa_canvas_config_set_dither_mode:
+ * @config: A #ChafaCanvasConfig
+ * @dither_mode: A #ChafaDitherMode
+ *
+ * Sets @config's stored #ChafaDitherMode to @dither_mode.
+ *
+ * Since: 1.2
+ **/
+void
+chafa_canvas_config_set_dither_mode (ChafaCanvasConfig *config, ChafaDitherMode dither_mode)
+{
+    g_return_if_fail (config != NULL);
+    g_return_if_fail (config->refs > 0);
+    g_return_if_fail (dither_mode < CHAFA_DITHER_MODE_MAX);
+
+    config->dither_mode = dither_mode;
+}
+
+/**
+ * chafa_canvas_config_get_dither_grain_size:
+ * @config: A #ChafaCanvasConfig
+ * @width_out: Pointer to a location to store grain width
+ * @height_out: Pointer to a location to store grain height
+ *
+ * Returns @config's dither grain size in @width_out and @height_out.
+ *
+ * Since: 1.2
+ **/
+void
+chafa_canvas_config_get_dither_grain_size (const ChafaCanvasConfig *config, gint *width_out, gint *height_out)
+{
+    g_return_if_fail (config != NULL);
+    g_return_if_fail (config->refs > 0);
+
+    if (width_out)
+        *width_out = config->dither_grain_width;
+    if (height_out)
+        *height_out = config->dither_grain_height;
+}
+
+/**
+ * chafa_canvas_config_set_dither_grain_size:
+ * @config: A #ChafaCanvasConfig
+ * @width: The desired grain width (1, 2, 4 or 8)
+ * @height: The desired grain height (1, 2, 4 or 8)
+ *
+ * Sets @config's stored dither grain size to @width by @height pixels. These
+ * values can be 1, 2, 4 or 8. 8 corresponds to the size of an entire
+ * character cell. The default is 4 pixels by 4 pixels.
+ *
+ * Since: 1.2
+ **/
+void
+chafa_canvas_config_set_dither_grain_size (ChafaCanvasConfig *config, gint width, gint height)
+{
+    g_return_if_fail (config != NULL);
+    g_return_if_fail (config->refs > 0);
+    g_return_if_fail (width == 1 || width == 2 || width == 4 || width == 8);
+    g_return_if_fail (height == 1 || height == 2 || height == 4 || height == 8);
+
+    config->dither_grain_width = width;
+    config->dither_grain_height = height;
+}
+
+/**
+ * chafa_canvas_config_get_dither_intensity:
+ * @config: A #ChafaCanvasConfig
+ *
+ * Returns the relative intensity of the dithering pattern applied during
+ * image conversion. 1.0 is the default, corresponding to a moderate
+ * intensity. 
+ *
+ * Returns: The relative dithering intensity
+ *
+ * Since: 1.2
+ **/
+gfloat
+chafa_canvas_config_get_dither_intensity (const ChafaCanvasConfig *config)
+{
+    g_return_val_if_fail (config != NULL, 1.0);
+    g_return_val_if_fail (config->refs > 0, 1.0);
+
+    return config->dither_intensity;
+}
+
+/**
+ * chafa_canvas_config_set_dither_intensity:
+ * @config: A #ChafaCanvasConfig
+ * @intensity: Desired relative dithering intensity
+ *
+ * Sets @config's stored relative intensity of the dithering pattern applied
+ * during image conversion. 1.0 is the default, corresponding to a moderate
+ * intensity. Possible values range from 0.0 to infinity, but in practice,
+ * values above 10.0 are rarely useful.
+ *
+ * Since: 1.2
+ **/
+void
+chafa_canvas_config_set_dither_intensity (ChafaCanvasConfig *config, gfloat intensity)
+{
+    g_return_if_fail (config != NULL);
+    g_return_if_fail (config->refs > 0);
+    g_return_if_fail (intensity >= 0.0);
+
+    config->dither_intensity = intensity;
+}
