@@ -22,6 +22,7 @@
 #include <math.h>
 #include <string.h>
 #include <glib.h>
+#include "smolscale/smolscale.h"
 #include "chafa/chafa.h"
 #include "chafa/chafa-private.h"
 
@@ -36,11 +37,28 @@
  * specify any parameters, like the geometry, color space and so on, you
  * must create a #ChafaCanvasConfig first.
  *
- * You can draw an image to the canvas using chafa_canvas_set_contents_rgba8 ()
+ * You can draw an image to the canvas using chafa_canvas_set_contents ()
  * and create an ANSI text representation of the canvas' current contents
  * using chafa_canvas_build_ansi ().
  **/
 
+/**
+ * ChafaPixelType:
+ * @CHAFA_PIXEL_RGBA8_PREMULTIPLIED: Premultiplied RGBA, 8 bits per channel.
+ * @CHAFA_PIXEL_BGRA8_PREMULTIPLIED: Premultiplied BGRA, 8 bits per channel.
+ * @CHAFA_PIXEL_ARGB8_PREMULTIPLIED: Premultiplied ARGB, 8 bits per channel.
+ * @CHAFA_PIXEL_ABGR8_PREMULTIPLIED: Premultiplied ABGR, 8 bits per channel.
+ * @CHAFA_PIXEL_RGBA8_UNASSOCIATED: Unassociated RGBA, 8 bits per channel.
+ * @CHAFA_PIXEL_BGRA8_UNASSOCIATED: Unassociated BGRA, 8 bits per channel.
+ * @CHAFA_PIXEL_ARGB8_UNASSOCIATED: Unassociated ARGB, 8 bits per channel.
+ * @CHAFA_PIXEL_ABGR8_UNASSOCIATED: Unassociated ABGR, 8 bits per channel.
+ * @CHAFA_PIXEL_RGB8: Packed RGB (no alpha), 8 bits per channel.
+ * @CHAFA_PIXEL_BGR8: Packed BGR (no alpha), 8 bits per channel.
+ * @CHAFA_PIXEL_MAX: Last supported pixel type, plus one.
+ *
+ * Pixel formats supported by #ChafaCanvas.
+ **/
+
 /* Fixed point multiplier */
 #define FIXED_MULT 16384
 
@@ -55,9 +73,22 @@
 #define INDEXED_16_CROP_PCT 5
 #define INDEXED_2_CROP_PCT  20
 
+/* Dithering */
+#define DITHER_GRAIN_WIDTH 4
+#define DITHER_GRAIN_HEIGHT 4
+#define DITHER_BASE_INTENSITY_FGBG 1.0
+#define DITHER_BASE_INTENSITY_16C  0.25
+#define DITHER_BASE_INTENSITY_256C 0.1
+#define BAYER_MATRIX_DIM_SHIFT 4
+#define BAYER_MATRIX_DIM (1 << (BAYER_MATRIX_DIM_SHIFT))
+#define BAYER_MATRIX_SIZE ((BAYER_MATRIX_DIM) * (BAYER_MATRIX_DIM))
+
 typedef struct
 {
     gint32 c [INTENSITY_MAX];
+
+    /* Lower and upper bounds */
+    gint32 min, max;
 }
 Histogram;
 
@@ -88,8 +119,14 @@ struct ChafaCanvas
     /* Used when setting pixel data */
     const guint8 *src_pixels;
     Histogram *hist;
+    ChafaPixelType src_pixel_type;
     gint src_width, src_height, src_rowstride;
     gint have_alpha_int;
+    gint dither_grain_width_shift, dither_grain_height_shift;
+
+    /* Set if we're doing bayer dithering */
+    gint *bayer_matrix;
+    gint bayer_size_shift;
 };
 
 typedef struct
@@ -790,20 +827,6 @@ update_cells (ChafaCanvas *canvas)
     g_thread_pool_free (thread_pool, FALSE, TRUE);
 }
 
-static void
-multiply_alpha (ChafaCanvas *canvas)
-{
-    ChafaPixel *p0, *p1;
-
-    p0 = canvas->pixels;
-    p1 = p0 + canvas->width_pixels * canvas->height_pixels;
-
-    for ( ; p0 < p1; p0++)
-    {
-        chafa_color_mix (&p0->col, &canvas->bg_color, &p0->col, 1000 - ((p0->col.ch [3] * 1000) / 255));
-    }
-}
-
 static gint
 rgb_to_intensity_fast (const ChafaColor *color)
 {
@@ -812,21 +835,48 @@ rgb_to_intensity_fast (const ChafaColor *color)
 }
 
 static void
-calc_histogram (ChafaCanvas *canvas, Histogram *hist, gint64 *accum)
+sum_histograms (const Histogram *hist_in, Histogram *hist_accum)
 {
-    ChafaPixel *p0, *p1;
+    gint i;
 
-    p0 = canvas->pixels;
-    p1 = p0 + canvas->width_pixels * canvas->height_pixels;
-
-    for (p0 = canvas->pixels; p0 < p1; p0++)
+    for (i = 0; i < INTENSITY_MAX; i++)
     {
-        gint v = rgb_to_intensity_fast (&p0->col);
-        *accum += v;
-        hist->c [v]++;
+        hist_accum->c [i] += hist_in->c [i];
     }
 }
 
+static void
+histogram_calc_bounds (ChafaCanvas *canvas, Histogram *hist, gint crop_pct)
+{
+    gint64 pixels_crop;
+    gint i;
+    gint t;
+
+    pixels_crop = (canvas->width_pixels * canvas->height_pixels * (((gint64) crop_pct * 1024) / 100)) / 1024;
+
+    /* Find lower bound */
+
+    for (i = 0, t = pixels_crop; i < INTENSITY_MAX; i++)
+    {
+        t -= hist->c [i];
+        if (t <= 0)
+            break;
+    }
+
+    hist->min = i;
+
+    /* Find upper bound */
+
+    for (i = INTENSITY_MAX - 1, t = pixels_crop; i >= 0; i--)
+    {
+        t -= hist->c [i];
+        if (t <= 0)
+            break;
+    }
+
+    hist->max = i;
+}
+
 static gint16
 normalize_ch (gint16 v, gint min, gint factor)
 {
@@ -843,46 +893,14 @@ normalize_ch (gint16 v, gint min, gint factor)
 }
 
 static void
-normalize_rgb (ChafaCanvas *canvas, gint crop_pct)
+normalize_rgb (ChafaCanvas *canvas, const Histogram *hist, gint dest_y, gint n_rows)
 {
-    Histogram hist;
-    gint64 accum = 0;
     ChafaPixel *p0, *p1;
-    gint min, max;
-    gint64 pixels_crop;
-    gint t;
     gint factor;
-    gint i;
-
-    memset (&hist, 0, sizeof (hist));
-    calc_histogram (canvas, &hist, &accum);
-    pixels_crop = (canvas->width_pixels * canvas->height_pixels * (((gint64) crop_pct * 1024) / 100)) / 1024;
-
-    /* Find lower bound */
-
-    for (i = 0, t = pixels_crop; i < INTENSITY_MAX; i++)
-    {
-        t -= hist.c [i];
-        if (t <= 0)
-            break;
-    }
-
-    min = i;
-
-    /* Find upper bound */
-
-    for (i = INTENSITY_MAX - 1, t = pixels_crop; i >= 0; i--)
-    {
-        t -= hist.c [i];
-        if (t <= 0)
-            break;
-    }
-
-    max = i;
 
     /* Make sure range is more or less sane */
 
-    if (min == max)
+    if (hist->min == hist->max)
         return;
 
 #if 0
@@ -894,20 +912,20 @@ normalize_rgb (ChafaCanvas *canvas, gint crop_pct)
 
     /* Adjust intensities */
 
-    factor = ((INTENSITY_MAX - 1) * FIXED_MULT) / (max - min);
+    factor = ((INTENSITY_MAX - 1) * FIXED_MULT) / (hist->max - hist->min);
 
 #if 0
     g_printerr ("[%d-%d] * %d, crop=%d     \n", min, max, factor, pixels_crop);
 #endif
 
-    p0 = canvas->pixels;
-    p1 = p0 + canvas->width_pixels * canvas->height_pixels;
+    p0 = canvas->pixels + dest_y * canvas->width_pixels;
+    p1 = p0 + n_rows * canvas->width_pixels;
 
-    for (p0 = canvas->pixels; p0 < p1; p0++)
+    for ( ; p0 < p1; p0++)
     {
-        p0->col.ch [0] = normalize_ch (p0->col.ch [0], min / 8, factor);
-        p0->col.ch [1] = normalize_ch (p0->col.ch [1], min / 8, factor);
-        p0->col.ch [2] = normalize_ch (p0->col.ch [2], min / 8, factor);
+        p0->col.ch [0] = normalize_ch (p0->col.ch [0], hist->min / 8, factor);
+        p0->col.ch [1] = normalize_ch (p0->col.ch [1], hist->min / 8, factor);
+        p0->col.ch [2] = normalize_ch (p0->col.ch [2], hist->min / 8, factor);
     }
 }
 
@@ -959,50 +977,97 @@ update_display_colors (ChafaCanvas *canvas)
 }
 
 static void
-rgba_to_internal_rgb (ChafaCanvas *canvas, gint dest_y, const guint8 *data, gint n_rows, gint rowstride)
+bayer_dither_pixel (ChafaCanvas *canvas, ChafaPixel *pixel, gint *matrix, gint x, gint y, guint size_shift, guint size_mask)
 {
-    ChafaPixel *pixel;
-    gint px, py;
-    gint x_inc, y_inc;
-    gint alpha_sum = 0;
+    gint bayer_index = (((y >> canvas->dither_grain_height_shift) & size_mask) << size_shift)
+        + ((x >> canvas->dither_grain_width_shift) & size_mask);
+    gint bayer_mod = matrix [bayer_index];
+    gint i;
 
-    x_inc = (canvas->src_width * FIXED_MULT) / (canvas->width_pixels);
-    y_inc = (canvas->src_height * FIXED_MULT) / (canvas->height_pixels);
-
-    pixel = canvas->pixels + dest_y * canvas->width_pixels;
-
-    for (py = dest_y; py < dest_y + n_rows; py++)
+    for (i = 0; i < 4; i++)
     {
-        const guint8 *data_row_p;
+        pixel->col.ch [i] += bayer_mod;
+        if (pixel->col.ch [i] < 0)
+            pixel->col.ch [i] = 0;
+        if (pixel->col.ch [i] > 255)
+            pixel->col.ch [i] = 255;
+    }
+}
 
-        data_row_p = data + ((py * y_inc) / FIXED_MULT) * rowstride;
+/* pixel must point to top-left pixel of the grain to be dithered */
+static void
+fs_dither_grain (ChafaCanvas *canvas, ChafaPixel *pixel, const ChafaPixel *error_in,
+                 ChafaPixel *error_out_0, ChafaPixel *error_out_1,
+                 ChafaPixel *error_out_2, ChafaPixel *error_out_3)
+{
+    gint grain_shift = canvas->dither_grain_width_shift + canvas->dither_grain_height_shift;
+    ChafaPixel next_error = { 0 };
+    ChafaPixel accum = { 0 };
+    ChafaColorCandidates cand = { 0 };
+    ChafaPixel *p;
+    const ChafaColor *c;
+    gint x, y, i;
 
-        for (px = 0; px < canvas->width_pixels; px++)
+    p = pixel;
+
+    for (y = 0; y < canvas->config.dither_grain_height; y++)
+    {
+        for (x = 0; x < canvas->config.dither_grain_width; x++, p++)
         {
-            const guint8 *data_p = data_row_p + ((px * x_inc) / FIXED_MULT) * 4;
-            ChafaColor col;
-
-            col.ch [0] = *data_p++;
-            col.ch [1] = *data_p++;
-            col.ch [2] = *data_p++;
-            col.ch [3] = *data_p;
-
-            alpha_sum += (0xff - col.ch [3]);
-
-            if (canvas->config.preprocessing_enabled
-                && canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16)
+            for (i = 0; i < 3; i++)
             {
-                boost_saturation_rgb (&col);
-                clamp_color_rgb (&col);
-            }
+                p->col.ch [i] += error_in->col.ch [i];
 
-            pixel->col = col;
-            pixel++;
+                if (canvas->config.color_space == CHAFA_COLOR_SPACE_RGB)
+                {
+                    if (p->col.ch [i] < 0)
+                    {
+                        next_error.col.ch [i] += p->col.ch [i];
+                        p->col.ch [i] = 0;
+                    }
+                    else if (p->col.ch [i] > 255)
+                    {
+                        next_error.col.ch [i] += p->col.ch [i] - 255;
+                        p->col.ch [i] = 255;
+                    }
+                }
+
+                accum.col.ch [i] += p->col.ch [i];
+            }
         }
+
+        p += canvas->width_pixels - canvas->config.dither_grain_width;
     }
 
-    if (alpha_sum > 0)
-        g_atomic_int_set (&canvas->have_alpha_int, 1);
+    for (i = 0; i < 3; i++)
+        accum.col.ch [i] >>= grain_shift;
+
+    /* Don't try to dither alpha */
+    accum.col.ch [3] = 0xff;
+
+    if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_256)
+        chafa_pick_color_256 (&accum.col, canvas->config.color_space, &cand);
+    else if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_240)
+        chafa_pick_color_240 (&accum.col, canvas->config.color_space, &cand);
+    else if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16)
+        chafa_pick_color_16 (&accum.col, canvas->config.color_space, &cand);
+    else
+        chafa_pick_color_fgbg (&accum.col, canvas->config.color_space,
+                               &canvas->fg_color, &canvas->bg_color, &cand);
+
+    c = chafa_get_palette_color_256 (cand.index [0], canvas->config.color_space);
+
+    for (i = 0; i < 3; i++)
+    {
+        /* FIXME: Floating point op is slow. Factor this out and make
+         * dither_intensity == 1.0 the fast path. */
+        next_error.col.ch [i] = ((next_error.col.ch [i] >> grain_shift) + (accum.col.ch [i] - c->ch [i]) * canvas->config.dither_intensity);
+
+        error_out_0->col.ch [i] += next_error.col.ch [i] * 7 / 16;
+        error_out_1->col.ch [i] += next_error.col.ch [i] * 1 / 16;
+        error_out_2->col.ch [i] += next_error.col.ch [i] * 5 / 16;
+        error_out_3->col.ch [i] += next_error.col.ch [i] * 3 / 16;
+    }
 }
 
 static void
@@ -1011,12 +1076,151 @@ convert_rgb_to_din99d (ChafaCanvas *canvas, gint dest_y, gint n_rows)
     ChafaPixel *pixel = canvas->pixels + dest_y * canvas->width_pixels;
     ChafaPixel *pixel_max = pixel + n_rows * canvas->width_pixels;
 
+    /* RGB -> DIN99d */
+
     for ( ; pixel < pixel_max; pixel++)
     {
         chafa_color_rgb_to_din99d (&pixel->col, &pixel->col);
     }
 }
 
+static void
+bayer_dither (ChafaCanvas *canvas, gint dest_y, gint n_rows)
+{
+    ChafaPixel *pixel = canvas->pixels + dest_y * canvas->width_pixels;
+    ChafaPixel *pixel_max = pixel + n_rows * canvas->width_pixels;
+    guint bayer_size_mask = (1 << canvas->bayer_size_shift) - 1;
+    gint x, y;
+
+    for (y = dest_y; pixel < pixel_max; y++)
+    {
+        for (x = 0; x < canvas->width_pixels; x++)
+        {
+            bayer_dither_pixel (canvas, pixel, canvas->bayer_matrix, x, y,
+                                canvas->bayer_size_shift, bayer_size_mask);
+            pixel++;
+        }
+    }
+}
+
+static void
+fs_dither (ChafaCanvas *canvas, gint dest_y, gint n_rows)
+{
+    ChafaPixel *pixel;
+    ChafaPixel *error_rows;
+    ChafaPixel *error_row [2];
+    ChafaPixel *pp;
+    gint width_grains = canvas->width_pixels >> canvas->dither_grain_width_shift;
+    gint x, y;
+
+    g_assert (canvas->width_pixels % canvas->config.dither_grain_width == 0);
+    g_assert (dest_y % canvas->config.dither_grain_height == 0);
+    g_assert (n_rows % canvas->config.dither_grain_height == 0);
+
+    dest_y >>= canvas->dither_grain_height_shift;
+    n_rows >>= canvas->dither_grain_height_shift;
+
+    error_rows = alloca (width_grains * 2 * sizeof (ChafaPixel));
+    error_row [0] = error_rows;
+    error_row [1] = error_rows + width_grains;
+
+    memset (error_row [0], 0, width_grains * sizeof (ChafaPixel));
+
+    for (y = dest_y; y < dest_y + n_rows; y++)
+    {
+        memset (error_row [1], 0, width_grains * sizeof (ChafaPixel));
+
+        if (!(y & 1))
+        {
+            /* Forwards pass */
+            pixel = canvas->pixels + (y << canvas->dither_grain_height_shift) * canvas->width_pixels;
+
+            fs_dither_grain (canvas, pixel, error_row [0],
+                             error_row [0] + 1,
+                             error_row [1] + 1,
+                             error_row [1],
+                             error_row [1] + 1);
+            pixel += canvas->config.dither_grain_width;
+
+            for (x = 1; ((x + 1) << canvas->dither_grain_width_shift) < canvas->width_pixels; x++)
+            {
+                fs_dither_grain (canvas, pixel, error_row [0] + x,
+                                 error_row [0] + x + 1,
+                                 error_row [1] + x + 1,
+                                 error_row [1] + x,
+                                 error_row [1] + x - 1);
+                pixel += canvas->config.dither_grain_width;
+            }
+
+            fs_dither_grain (canvas, pixel, error_row [0] + x,
+                             error_row [1] + x,
+                             error_row [1] + x,
+                             error_row [1] + x - 1,
+                             error_row [1] + x - 1);
+        }
+        else
+        {
+            /* Backwards pass */
+            pixel = canvas->pixels + (y << canvas->dither_grain_height_shift) * canvas->width_pixels + canvas->width_pixels - canvas->config.dither_grain_width;
+
+            fs_dither_grain (canvas, pixel, error_row [0] + width_grains - 1,
+                             error_row [0] + width_grains - 2,
+                             error_row [1] + width_grains - 2,
+                             error_row [1] + width_grains - 1,
+                             error_row [1] + width_grains - 2);
+
+            pixel -= canvas->config.dither_grain_width;
+
+            for (x = width_grains - 2; x > 0; x--)
+            {
+                fs_dither_grain (canvas, pixel, error_row [0] + x,
+                                 error_row [0] + x - 1,
+                                 error_row [1] + x - 1,
+                                 error_row [1] + x,
+                                 error_row [1] + x + 1);
+                pixel -= canvas->config.dither_grain_width;
+            }
+
+            fs_dither_grain (canvas, pixel, error_row [0],
+                             error_row [1],
+                             error_row [1],
+                             error_row [1] + 1,
+                             error_row [1] + 1);
+        }
+
+        pp = error_row [0];
+        error_row [0] = error_row [1];
+        error_row [1] = pp;
+    }
+}
+
+static void
+bayer_and_convert_rgb_to_din99d (ChafaCanvas *canvas, gint dest_y, gint n_rows)
+{
+    ChafaPixel *pixel = canvas->pixels + dest_y * canvas->width_pixels;
+    ChafaPixel *pixel_max = pixel + n_rows * canvas->width_pixels;
+    guint bayer_size_mask = (1 << canvas->bayer_size_shift) - 1;
+    gint x, y;
+
+    for (y = dest_y; pixel < pixel_max; y++)
+    {
+        for (x = 0; x < canvas->width_pixels; x++)
+        {
+            bayer_dither_pixel (canvas, pixel, canvas->bayer_matrix, x, y,
+                                canvas->bayer_size_shift, bayer_size_mask);
+            chafa_color_rgb_to_din99d (&pixel->col, &pixel->col);
+            pixel++;
+        }
+    }
+}
+
+static void
+fs_and_convert_rgb_to_din99d (ChafaCanvas *canvas, gint dest_y, gint n_rows)
+{
+    convert_rgb_to_din99d (canvas, dest_y, n_rows);
+    fs_dither (canvas, dest_y, n_rows);
+}
+
 static void
 maybe_clear (ChafaCanvas *canvas)
 {
@@ -1034,99 +1238,348 @@ maybe_clear (ChafaCanvas *canvas)
     }
 }
 
-#define ROWS_PER_BATCH 8
+typedef struct
+{
+    ChafaCanvas *canvas;
+    Histogram hist;
+    gint n_batches_pixels;
+    gint n_rows_per_batch_pixels;
+    gint n_batches_cells;
+    gint n_rows_per_batch_cells;
+    SmolScaleCtx *scale_ctx;
+}
+PrepareContext;
+
+typedef struct
+{
+    gint first_row;
+    gint n_rows;
+    Histogram hist;
+}
+PreparePixelsBatch1;
+
+static void
+prepare_pixels_1_inner (PreparePixelsBatch1 *work,
+                        PrepareContext *prep_ctx,
+                        const guint8 *data_p,
+                        ChafaPixel *pixel_out,
+                        gint *alpha_sum)
+{
+    ChafaColor *col = &pixel_out->col;
+    gint v;
+
+    col->ch [0] = data_p [0];
+    col->ch [1] = data_p [1];
+    col->ch [2] = data_p [2];
+    col->ch [3] = data_p [3];
+
+    *alpha_sum += (0xff - col->ch [3]);
+
+    if (prep_ctx->canvas->config.preprocessing_enabled
+        && prep_ctx->canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16)
+    {
+        boost_saturation_rgb (col);
+        clamp_color_rgb (col);
+    }
+
+    /* Build histogram */
+    v = rgb_to_intensity_fast (col);
+    work->hist.c [v]++;
+}
+
+static void
+prepare_pixels_1_worker_nearest (PreparePixelsBatch1 *work, PrepareContext *prep_ctx)
+{
+    ChafaPixel *pixel;
+    ChafaCanvas *canvas;
+    gint dest_y;
+    gint px, py;
+    gint x_inc, y_inc;
+    gint alpha_sum = 0;
+    const guint8 *data;
+    gint n_rows;
+    gint rowstride;
+
+    canvas = prep_ctx->canvas;
+    dest_y = work->first_row;
+    data = canvas->src_pixels;
+    n_rows = work->n_rows;
+    rowstride = canvas->src_rowstride;
+
+    x_inc = (canvas->src_width * FIXED_MULT) / (canvas->width_pixels);
+    y_inc = (canvas->src_height * FIXED_MULT) / (canvas->height_pixels);
+
+    pixel = canvas->pixels + dest_y * canvas->width_pixels;
+
+    for (py = dest_y; py < dest_y + n_rows; py++)
+    {
+        const guint8 *data_row_p;
+
+        data_row_p = data + ((py * y_inc) / FIXED_MULT) * rowstride;
+
+        for (px = 0; px < canvas->width_pixels; px++)
+        {
+            const guint8 *data_p = data_row_p + ((px * x_inc) / FIXED_MULT) * 4;
+            prepare_pixels_1_inner (work, prep_ctx, data_p, pixel++, &alpha_sum);
+        }
+    }
+
+    if (alpha_sum > 0)
+        g_atomic_int_set (&canvas->have_alpha_int, 1);
+}
+
+static void
+prepare_pixels_1_worker_smooth (PreparePixelsBatch1 *work, PrepareContext *prep_ctx)
+{
+    ChafaCanvas *canvas = prep_ctx->canvas;
+    ChafaPixel *pixel, *pixel_max;
+    gint alpha_sum = 0;
+    guint8 *scaled_data;
+    const guint8 *data_p;
+
+    canvas = prep_ctx->canvas;
+
+    scaled_data = g_malloc (canvas->width_pixels * work->n_rows * sizeof (guint32));
+    smol_scale_batch_full (prep_ctx->scale_ctx, scaled_data, work->first_row, work->n_rows);
+
+    data_p = scaled_data;
+    pixel = canvas->pixels + work->first_row * canvas->width_pixels;
+    pixel_max = pixel + work->n_rows * canvas->width_pixels;
+
+    while (pixel < pixel_max)
+    {
+        prepare_pixels_1_inner (work, prep_ctx, data_p, pixel++, &alpha_sum);
+        data_p += 4;
+    }
+
+    g_free (scaled_data);
+
+    if (alpha_sum > 0)
+        g_atomic_int_set (&canvas->have_alpha_int, 1);
+}
+
+static void
+prepare_pixels_pass_1 (PrepareContext *prep_ctx)
+{
+    GThreadPool *thread_pool;
+    PreparePixelsBatch1 *batches;
+    gint cy;
+    gint i;
+
+    /* First pass
+     * ----------
+     *
+     * - Scale and convert pixel format
+     * - Apply local preprocessing like saturation boost (optional)
+     * - Generate histogram for later passes (e.g. for normalization)
+     * - Figure out if we have alpha transparency
+     */
+
+    batches = g_new0 (PreparePixelsBatch1, prep_ctx->n_batches_pixels);
+
+    thread_pool = g_thread_pool_new ((GFunc) ((prep_ctx->canvas->work_factor_int < 3
+                                               && prep_ctx->canvas->src_pixel_type == CHAFA_PIXEL_RGBA8_UNASSOCIATED)
+                                              ? prepare_pixels_1_worker_nearest
+                                              : prepare_pixels_1_worker_smooth),
+                                     prep_ctx,
+                                     g_get_num_processors (),
+                                     FALSE,
+                                     NULL);
+
+    for (cy = 0, i = 0;
+         cy < prep_ctx->canvas->height_pixels;
+         cy += prep_ctx->n_rows_per_batch_pixels, i++)
+    {
+        PreparePixelsBatch1 *batch = &batches [i];
+
+        batch->first_row = cy;
+        batch->n_rows = MIN (prep_ctx->canvas->height_pixels - cy, prep_ctx->n_rows_per_batch_pixels);
+
+        g_thread_pool_push (thread_pool, batch, NULL);
+    }
+
+    /* Wait for threads to finish */
+    g_thread_pool_free (thread_pool, FALSE, TRUE);
+
+    /* Generate final histogram */
+    if (prep_ctx->canvas->config.preprocessing_enabled)
+    {
+        for (i = 0; i < prep_ctx->n_batches_pixels; i++)
+            sum_histograms (&batches [i].hist, &prep_ctx->hist);
+
+        histogram_calc_bounds (prep_ctx->canvas, &prep_ctx->hist,
+                               prep_ctx->canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16 ? INDEXED_16_CROP_PCT : INDEXED_2_CROP_PCT);
+    }
+
+    /* Report alpha situation */
+    if (prep_ctx->canvas->have_alpha_int)
+        prep_ctx->canvas->have_alpha = TRUE;
+
+    g_free (batches);
+}
 
 typedef struct
 {
     gint first_row;
     gint n_rows;
 }
-PreparePixelsWork;
+PreparePixelsBatch2;
 
 static void
-prepare_pixels_1_worker (PreparePixelsWork *work, ChafaCanvas *canvas)
+composite_alpha_on_bg (ChafaCanvas *canvas, gint first_row, gint n_rows)
 {
-    rgba_to_internal_rgb (canvas,
-                          work->first_row,
-                          canvas->src_pixels,
-                          work->n_rows,
-                          canvas->src_rowstride);
+    ChafaPixel *p0, *p1;
 
-    g_slice_free (PreparePixelsWork, work);
+    p0 = canvas->pixels + first_row * canvas->width_pixels;
+    p1 = p0 + n_rows * canvas->width_pixels;
+
+    for ( ; p0 < p1; p0++)
+    {
+        p0->col.ch [0] += (canvas->bg_color.ch [0] * (255 - (guint32) p0->col.ch [3])) / 255;
+        p0->col.ch [1] += (canvas->bg_color.ch [1] * (255 - (guint32) p0->col.ch [3])) / 255;
+        p0->col.ch [2] += (canvas->bg_color.ch [2] * (255 - (guint32) p0->col.ch [3])) / 255;
+    }
 }
 
 static void
-prepare_pixels_3_worker (PreparePixelsWork *work, ChafaCanvas *canvas)
+prepare_pixels_2_worker (PreparePixelsBatch2 *work, PrepareContext *prep_ctx)
 {
-    convert_rgb_to_din99d (canvas,
-                           work->first_row,
-                           work->n_rows);
+    ChafaCanvas *canvas = prep_ctx->canvas;
 
-    g_slice_free (PreparePixelsWork, work);
+    if (canvas->config.preprocessing_enabled
+        && (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16
+            || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG_BGFG
+            || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG))
+        normalize_rgb (canvas, &prep_ctx->hist, work->first_row, work->n_rows);
+
+    if (canvas->config.color_space == CHAFA_COLOR_SPACE_DIN99D)
+    {
+        if (canvas->config.dither_mode == CHAFA_DITHER_MODE_ORDERED)
+        {
+            bayer_and_convert_rgb_to_din99d (canvas,
+                                             work->first_row,
+                                             work->n_rows);
+        }
+        else if (canvas->config.dither_mode == CHAFA_DITHER_MODE_DIFFUSION)
+        {
+            fs_and_convert_rgb_to_din99d (canvas,
+                                          work->first_row,
+                                          work->n_rows);
+        }
+        else
+        {
+            convert_rgb_to_din99d (canvas,
+                                   work->first_row,
+                                   work->n_rows);
+        }
+    }
+    else if (canvas->config.dither_mode == CHAFA_DITHER_MODE_ORDERED)
+    {
+        bayer_dither (canvas,
+                      work->first_row,
+                      work->n_rows);
+    }
+    else if (canvas->config.dither_mode == CHAFA_DITHER_MODE_DIFFUSION)
+    {
+        fs_dither (canvas,
+                   work->first_row,
+                   work->n_rows);
+    }
+
+    /* Must do this after DIN99d conversion, since bg_color will be DIN99d too */
+    if (canvas->have_alpha)
+        composite_alpha_on_bg (canvas, work->first_row, work->n_rows);
+}
+
+static gboolean
+need_pass_2 (ChafaCanvas *canvas)
+{
+    if ((canvas->config.preprocessing_enabled
+         && (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16
+             || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG_BGFG
+             || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG))
+        || canvas->have_alpha
+        || canvas->config.color_space == CHAFA_COLOR_SPACE_DIN99D
+        || canvas->config.dither_mode != CHAFA_DITHER_MODE_NONE)
+        return TRUE;
+
+    return FALSE;
 }
 
 static void
-prepare_pixel_data (ChafaCanvas *canvas)
+prepare_pixels_pass_2 (PrepareContext *prep_ctx)
 {
     GThreadPool *thread_pool;
+    PreparePixelsBatch1 *batches;
     gint cy;
+    gint i;
 
-    /* First pass -- convert pixel format and apply local preprocessing
-     * like saturation boost */
+    /* Second pass
+     * -----------
+     *
+     * - Normalization (optional)
+     * - Dithering (optional)
+     * - Color space conversion; DIN99d (optional)
+     */
 
-    thread_pool = g_thread_pool_new ((GFunc) prepare_pixels_1_worker,
-                                     canvas,
+    if (!need_pass_2 (prep_ctx->canvas))
+        return;
+
+    batches = g_new0 (PreparePixelsBatch1, prep_ctx->n_batches_pixels);
+
+    thread_pool = g_thread_pool_new ((GFunc) prepare_pixels_2_worker,
+                                     prep_ctx,
                                      g_get_num_processors (),
                                      FALSE,
                                      NULL);
 
-    for (cy = 0; cy < canvas->height_pixels; cy += ROWS_PER_BATCH)
+    for (cy = 0, i = 0;
+         cy < prep_ctx->canvas->height_pixels;
+         cy += prep_ctx->n_rows_per_batch_pixels, i++)
     {
-        PreparePixelsWork *work = g_slice_new (PreparePixelsWork);
+        PreparePixelsBatch1 *batch = &batches [i];
 
-        work->first_row = cy;
-        work->n_rows = MIN (canvas->height_pixels - cy, ROWS_PER_BATCH);
+        batch->first_row = cy;
+        batch->n_rows = MIN (prep_ctx->canvas->height_pixels - cy, prep_ctx->n_rows_per_batch_pixels);
 
-        g_thread_pool_push (thread_pool, work, NULL);
+        g_thread_pool_push (thread_pool, batch, NULL);
     }
 
     /* Wait for threads to finish */
     g_thread_pool_free (thread_pool, FALSE, TRUE);
 
-    /* Second pass -- apply global preprocessing */
+    g_free (batches);
+}
 
-    if (canvas->config.preprocessing_enabled)
-    {
-        if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_INDEXED_16)
-            normalize_rgb (canvas, INDEXED_16_CROP_PCT);
-        else if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG_BGFG
-                 || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG)
-            normalize_rgb (canvas, INDEXED_2_CROP_PCT);
-    }
+static void
+prepare_pixel_data (ChafaCanvas *canvas)
+{
+    PrepareContext prep_ctx = { 0 };
+    guint n_cpus;
 
-    /* Third pass -- convert to DIN99d */
+    n_cpus = g_get_num_processors ();
 
-    if (canvas->config.color_space == CHAFA_COLOR_SPACE_DIN99D)
-    {
-        thread_pool = g_thread_pool_new ((GFunc) prepare_pixels_3_worker,
-                                         canvas,
-                                         g_get_num_processors (),
-                                         FALSE,
-                                         NULL);
+    prep_ctx.canvas = canvas;
+    prep_ctx.n_batches_pixels = (canvas->height_pixels + n_cpus - 1) / n_cpus;
+    prep_ctx.n_rows_per_batch_pixels = (canvas->height_pixels + prep_ctx.n_batches_pixels - 1) / prep_ctx.n_batches_pixels;
+    prep_ctx.n_batches_cells = (canvas->config.height + n_cpus - 1) / n_cpus;
+    prep_ctx.n_rows_per_batch_cells = (canvas->config.height + prep_ctx.n_batches_cells - 1) / prep_ctx.n_batches_cells;
 
-        for (cy = 0; cy < canvas->height_pixels; cy += ROWS_PER_BATCH)
-        {
-            PreparePixelsWork *work = g_slice_new (PreparePixelsWork);
+    prep_ctx.scale_ctx = smol_scale_new ((SmolPixelType) canvas->src_pixel_type,
+                                         (const guint32 *) canvas->src_pixels,
+                                         canvas->src_width,
+                                         canvas->src_height,
+                                         canvas->src_rowstride,
+                                         SMOL_PIXEL_RGBA8_PREMULTIPLIED,
+                                         NULL,
+                                         canvas->width_pixels,
+                                         canvas->height_pixels,
+                                         canvas->width_pixels * sizeof (guint32));
 
-            work->first_row = cy;
-            work->n_rows = MIN (canvas->height_pixels - cy, ROWS_PER_BATCH);
+    prepare_pixels_pass_1 (&prep_ctx);
+    prepare_pixels_pass_2 (&prep_ctx);
 
-            g_thread_pool_push (thread_pool, work, NULL);
-        }
-
-        /* Wait for threads to finish */
-        g_thread_pool_free (thread_pool, FALSE, TRUE);
-    }
+    smol_scale_destroy (prep_ctx.scale_ctx);
 }
 
 static void
@@ -1347,6 +1800,26 @@ build_ansi_gstring (ChafaCanvas *canvas)
     return gs;
 }
 
+static gint
+calc_dither_grain_shift (gint size)
+{
+    switch (size)
+    {
+        case 1:
+            return 0;
+        case 2:
+            return 1;
+        case 4:
+            return 2;
+        case 8:
+            return 3;
+        default:
+            g_assert_not_reached ();
+    }
+
+    return 0;
+}
+
 /**
  * chafa_canvas_new:
  * @config: Configuration to use or %NULL for hardcoded defaults
@@ -1392,9 +1865,43 @@ chafa_canvas_new (const ChafaCanvasConfig *config)
     /* In truecolor mode we don't support any fancy color spaces for now, since
      * we'd have to convert back to RGB space when emitting control codes, and
      * the code for that has yet to be written. In palette modes we just use
-     * the palette mappings. */
+     * the palette mappings.
+     *
+     * There is also no reason to dither in truecolor mode. */
     if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_TRUECOLOR)
+    {
         canvas->config.color_space = CHAFA_COLOR_SPACE_RGB;
+        canvas->config.dither_mode = CHAFA_DITHER_MODE_NONE;
+    }
+
+    canvas->dither_grain_width_shift = calc_dither_grain_shift (canvas->config.dither_grain_width);
+    canvas->dither_grain_height_shift = calc_dither_grain_shift (canvas->config.dither_grain_height);
+
+    if (canvas->config.dither_mode == CHAFA_DITHER_MODE_ORDERED)
+    {
+        gdouble dither_intensity;
+
+        switch (canvas->config.canvas_mode)
+        {
+            case CHAFA_CANVAS_MODE_INDEXED_256:
+            case CHAFA_CANVAS_MODE_INDEXED_240:
+                dither_intensity = DITHER_BASE_INTENSITY_256C;
+                break;
+            case CHAFA_CANVAS_MODE_INDEXED_16:
+                dither_intensity = DITHER_BASE_INTENSITY_16C;
+                break;
+            case CHAFA_CANVAS_MODE_FGBG:
+            case CHAFA_CANVAS_MODE_FGBG_BGFG:
+                dither_intensity = DITHER_BASE_INTENSITY_FGBG;
+                break;
+            default:
+                g_assert_not_reached ();
+                break;
+        }
+
+        canvas->bayer_size_shift = BAYER_MATRIX_DIM_SHIFT;
+        canvas->bayer_matrix = chafa_gen_bayer_matrix (BAYER_MATRIX_DIM, dither_intensity * canvas->config.dither_intensity);
+    }
 
     update_display_colors (canvas);
 
@@ -1425,6 +1932,7 @@ chafa_canvas_new_similar (ChafaCanvas *orig)
     canvas->pixels = NULL;
     canvas->cells = g_new (ChafaCanvasCell, canvas->config.width * canvas->config.height);
     canvas->needs_clear = TRUE;
+    canvas->bayer_matrix = g_memdup (orig->bayer_matrix, BAYER_MATRIX_SIZE * sizeof (gint));
 
     return canvas;
 }
@@ -1468,6 +1976,7 @@ chafa_canvas_unref (ChafaCanvas *canvas)
         chafa_canvas_config_deinit (&canvas->config);
         g_free (canvas->pixels);
         g_free (canvas->cells);
+        g_free (canvas->bayer_matrix);
         g_free (canvas);
     }
 }
@@ -1491,6 +2000,58 @@ chafa_canvas_peek_config (ChafaCanvas *canvas)
     return &canvas->config;
 }
 
+/**
+ * chafa_canvas_draw_all_pixels:
+ * @canvas: Canvas whose pixel data to replace
+ * @src_pixel_type: Pixel format of @src_pixels
+ * @src_pixels: Pointer to the start of source pixel memory
+ * @src_width: Width in pixels of source pixel data
+ * @src_height: Height in pixels of source pixel data
+ * @src_rowstride: Number of bytes between the start of each pixel row
+ *
+ * Replaces pixel data of @canvas with a copy of that found at @src_pixels,
+ * which must be in one of the formats supported by #ChafaPixelType.
+ *
+ * Since: 1.2
+ **/
+void
+chafa_canvas_draw_all_pixels (ChafaCanvas *canvas, ChafaPixelType src_pixel_type,
+                              const guint8 *src_pixels,
+                              gint src_width, gint src_height, gint src_rowstride)
+{
+    g_return_if_fail (canvas != NULL);
+    g_return_if_fail (canvas->refs > 0);
+    g_return_if_fail (src_pixel_type < CHAFA_PIXEL_MAX);
+    g_return_if_fail (src_pixels != NULL);
+    g_return_if_fail (src_width >= 0);
+    g_return_if_fail (src_height >= 0);
+
+    if (src_width == 0 || src_height == 0)
+        return;
+
+    canvas->pixels = g_new (ChafaPixel, canvas->width_pixels * canvas->height_pixels);
+    canvas->hist = g_new (Histogram, 1);
+
+    canvas->src_pixel_type = src_pixel_type;
+    canvas->src_pixels = src_pixels;
+    canvas->src_width = src_width;
+    canvas->src_height = src_height;
+    canvas->src_rowstride = src_rowstride;
+    canvas->have_alpha_int = 0;
+
+    prepare_pixel_data (canvas);
+
+    if (canvas->config.alpha_threshold == 0)
+        canvas->have_alpha = FALSE;
+
+    update_cells (canvas);
+    canvas->needs_clear = FALSE;
+
+    g_free (canvas->hist);
+    g_free (canvas->pixels);
+    canvas->pixels = NULL;
+}
+
 /**
  * chafa_canvas_set_contents_rgba8:
  * @canvas: Canvas whose pixel data to replace
@@ -1503,46 +2064,15 @@ chafa_canvas_peek_config (ChafaCanvas *canvas)
  * The source data must be in packed 8-bits-per-channel RGBA format. The
  * alpha value is expressed as opacity (0xff is opaque) and is not
  * premultiplied.
+ *
+ * Deprecated: 1.2: Use chafa_canvas_draw_all_pixels() instead.
  **/
 void
 chafa_canvas_set_contents_rgba8 (ChafaCanvas *canvas, const guint8 *src_pixels,
                                  gint src_width, gint src_height, gint src_rowstride)
 {
-    g_return_if_fail (canvas != NULL);
-    g_return_if_fail (canvas->refs > 0);
-    g_return_if_fail (src_pixels != NULL);
-    g_return_if_fail (src_width >= 0);
-    g_return_if_fail (src_height >= 0);
-
-    if (src_width == 0 || src_height == 0)
-        return;
-
-    canvas->pixels = g_new (ChafaPixel, canvas->width_pixels * canvas->height_pixels);
-    canvas->hist = g_new (Histogram, 1);
-
-    canvas->src_pixels = src_pixels;
-    canvas->src_width = src_width;
-    canvas->src_height = src_height;
-    canvas->src_rowstride = src_rowstride;
-    canvas->have_alpha_int = 0;
-
-    prepare_pixel_data (canvas);
-
-    if (canvas->have_alpha_int)
-        canvas->have_alpha = TRUE;
-
-    if (canvas->have_alpha)
-        multiply_alpha (canvas);
-
-    if (canvas->config.alpha_threshold == 0)
-        canvas->have_alpha = FALSE;
-
-    update_cells (canvas);
-    canvas->needs_clear = FALSE;
-
-    g_free (canvas->hist);
-    g_free (canvas->pixels);
-    canvas->pixels = NULL;
+    chafa_canvas_draw_all_pixels (canvas, CHAFA_PIXEL_RGBA8_UNASSOCIATED,
+                                  src_pixels, src_width, src_height, src_rowstride);
 }
 
 /**
