@@ -1641,58 +1641,6 @@ static const ChafaSymbolDef symbol_defs [] =
         "        "
         "        "
     },
-    {
-        /* Variant */
-        CHAFA_SYMBOL_TAG_DOT,
-        0x25aa, /* Black small square */
-        "        "
-        "        "
-        " XXXX   "
-        " XXXX   "
-        " XXXX   "
-        " XXXX   "
-        "        "
-        "        "
-    },
-    {
-        /* Variant */
-        CHAFA_SYMBOL_TAG_DOT,
-        0x25aa, /* Black small square */
-        "        "
-        "        "
-        "   XXXX "
-        "   XXXX "
-        "   XXXX "
-        "   XXXX "
-        "        "
-        "        "
-    },
-    {
-        /* Variant */
-        CHAFA_SYMBOL_TAG_DOT,
-        0x25aa, /* Black small square */
-        "        "
-        "  XXXX  "
-        "  XXXX  "
-        "  XXXX  "
-        "  XXXX  "
-        "        "
-        "        "
-        "        "
-    },
-    {
-        /* Variant */
-        CHAFA_SYMBOL_TAG_DOT,
-        0x25aa, /* Black small square */
-        "        "
-        "        "
-        "        "
-        "  XXXX  "
-        "  XXXX  "
-        "  XXXX  "
-        "  XXXX  "
-        "        "
-    },
     {
         /* Black up-pointing triangle */
         CHAFA_SYMBOL_TAG_GEOMETRIC,
@@ -1940,6 +1888,19 @@ static const ChafaSymbolDef symbol_defs [] =
         " XXX XXX"
         "XX XXX X"
     },
+    {
+        /* Greek Capital Letter Xi */
+        CHAFA_SYMBOL_TAG_EXTRA,
+        0x039e,
+        "        "
+        " XXXXXX "
+        "        "
+        "  XXXX  "
+        "        "
+        " XXXXXX "
+        "        "
+        "        "
+    },
     {
         0, 0, ""
     }
